--- Box2D plugin for STI
-- @module box2d
-- @author Landon Manning
-- @author Claude Marinier
-- @copyright 2021
-- @license MIT/X11

local love  = _G.love
local utils = require((...):gsub('plugins.box2d', 'utils'))
local lg    = require((...):gsub('plugins.box2d', 'graphics'))

return {
	box2d_LICENSE     = "MIT/X11",
	box2d_URL         = "https://github.com/karai17/Simple-Tiled-Implementation",
	box2d_VERSION     = "2.3.2.7",
	box2d_DESCRIPTION = "Box2D hooks for STI.",

	--- Initialize Box2D physics world.
	-- @param world The Box2D world to add objects to.
    --
	box2d_init = function(map, world)

		assert(love.physics, "To use the Box2D plugin, enable the love.physics module.")

		local body      = love.physics.newBody(world, map.offsetx, map.offsety)
		local collision = {
			body = body,
		}

        local function tableIsEmpty( aTable )
            -- recommended for efficiency, not sure it is needed
            local next = next  
            return next(aTable) == nil
            -- another suggestion
--          return type(next(aTable)) == "nil"
        end

        local function dumpTable(o)
            if type(o) == 'table' then
                local s = '{ '
                for k,v in pairs(o) do
                    if type(k) ~= 'number' then k = '"'..k..'"' end
                    s = s .. '['..k..'] = ' .. dumpTable(v) .. ','
                end
                return s .. '} '
            else
                return tostring(o)
            end
        end

        local function dumpTableFew(o,max,depth)
            max = max or 2
            depth = depth or 1
            if type(o) == 'table' and depth < max then
                local s = '{ '
                for k,v in pairs(o) do
                    if type(k) ~= 'number' then k = '"'..k..'"' end
                    s = s .. '['..k..'] = ' .. dumpTableFew(v,max,depth+1) .. ','
                end
                return s .. '} '
            else
                return tostring(o)
            end
        end

        -- utility function
        --
		local function getPolygonVertices(object)
			local vertices = {}
			for _, vertex in ipairs(object.polygon) do
				table.insert(vertices, vertex.x)
				table.insert(vertices, vertex.y)
			end
			return vertices
		end

        -- Set some custom properties from userdata
        -- (or use default set by box2d)
        --
        local function setFixtureProperties(fixture, userdata)
			fixture:setUserData(userdata)
			fixture:setFriction(userdata.properties.friction       or 0.2)
			fixture:setRestitution(userdata.properties.restitution or 0.0)
			fixture:setSensor(userdata.properties.sensor           or false)
			fixture:setFilterData(
				userdata.properties.categories or 1,
				userdata.properties.mask       or 65535,
				userdata.properties.group      or 0
			)
        end

        -- create a new body for a collidable object which is not a tile
        --
        -- collidable tiles should be static and attached to the main body
        -- (the map)
        --
        -- After an object is converted to triangles, all the triangles
        -- must be kept together. How can I do this?
        --
		local function addObjectToWorld(obj, userdata, tileOrObject)

            -- the tileOrObject parameter is not used,
            -- it is stored in the collision table

            print( "\naddObjectToWorld" )
            print( "  obj x     = " .. obj.x )
            print( "  obj y     = " .. obj.y )
            print( "  obj shape = " .. obj.shape )
            print( "  #vertices = " .. #obj.vertices )

            -- start with the body, there is only one
            --
            -- may attach many shapes to it
            --
			local currentBody

			-- dynamic is used for objects or players
            --
			if userdata.properties.dynamic == true then
                print("use new dynamic body")
				currentBody = love.physics.newBody(world, obj.x, obj.y, 'dynamic')

			-- static is used for things that shouldn't move, like walls/ground
            --
			elseif userdata.properties.static == true then
                print("use new static body")
				currentBody = love.physics.newBody(world, obj.x, obj.y, 'static')

			-- kinematic objects are static but effect other bodies
            --
			elseif userdata.properties.kinematic == true then
                print("use new kinematic body")
				currentBody = love.physics.newBody(world, obj.x, obj.y, 'kinematic')

            -- everything else is attached to the main body
            --
            else
                print("use main body")
			    currentBody = body
			end

            -- now shape and fixture
            --
            -- will trangulate polygons and attach them to the same body
            --
			local shape
			local fixture

            -- this code assumes that everything that is not a polyline
            -- has been converted to a polygon
            --
			if obj.shape == "polyline" then
                -- check the Box2D documentation: Collision, Shapes
                -- edges are two-sided
                -- chains are one-sided: watch vertex order (this would
                -- affect the way I draw collision shapes in Tiled)
                --
				shape = love.physics.newEdgeShape(unpack(obj.vertices))

                fixture = love.physics.newFixture(currentBody, shape)
                setFixtureProperties(fixture, userdata)

                -- add it to the collision table
                --
                table.insert(collision, {
                                            object  = tileOrObject,
                                            body    = currentBody,
                                            shape   = shape,
                                            fixture = fixture,
                                        })

--              -- original code
--              --
--              -- why is a four vertex polyline treated differently?
--              -- it may be to fix a specific problem
--				if #vertices == 4 then
--					shape = love.physics.newEdgeShape(unpack(vertices))
--				else
--					shape = love.physics.newChainShape(false, unpack(vertices))
--				end

            elseif obj.shape == "polygon" then
                -- These need to be converted to triangles
                --
                -- Elipses were converted to polygons in calculateObjectPosition
                --
				local triangles = love.math.triangulate(obj.vertices)

				for _, triangle in ipairs(triangles) do
                    -- make shape and fixture for each
                    --
                    shape = love.physics.newPolygonShape(unpack(triangle))

                    fixture = love.physics.newFixture(currentBody, shape)

                    setFixtureProperties(fixture, userdata)

                    -- add it to the collision table
                    --
                    table.insert(collision, {
                                                object  = tileOrObject,
                                                body    = currentBody,
                                                shape   = shape,
                                                fixture = fixture,
                                            })
				end

			else
				shape = love.physics.newPolygonShape(unpack(obj.vertices))
                fixture = love.physics.newFixture(currentBody, shape)
                setFixtureProperties(fixture, userdata)
                -- add it to the collision table
                --
                table.insert(collision, {
                                            object  = tileOrObject,
                                            body    = currentBody,
                                            shape   = shape,
                                            fixture = fixture,
                                        })
			end

--[[
            fixture = love.physics.newFixture(currentBody, shape)

            setFixtureProperties(fixture, userdata)

			-- Set some custom properties from userdata
            -- (or use default set by box2d)
            --
			fixture:setUserData(userdata)
			fixture:setFriction(userdata.properties.friction       or 0.2)
			fixture:setRestitution(userdata.properties.restitution or 0.0)
			fixture:setSensor(userdata.properties.sensor           or false)
			fixture:setFilterData(
				userdata.properties.categories or 1,
				userdata.properties.mask       or 65535,
				userdata.properties.group      or 0
			)

            -- add it to the collision table
            --
			local obj = {
				object  = object,
				body    = currentBody,
				shape   = shape,
				fixture = fixture,
			}
			table.insert(collision, obj)
--]]

		end

        -- there is a lot happening in this function
        --
		local function calculateObjectPosition(object, tile)

			local o = {
				shape   = object.shape,
				x       = (object.dx or object.x) + map.offsetx,
				y       = (object.dy or object.y) + map.offsety,
				w       = object.width,
				h       = object.height,
				polygon = object.polygon or object.polyline or object.ellipse or object.rectangle
			}

            print( "\ncalculateObjectPosition" )
            print( "  object:", dumpTableFew(object,2) )
            print( "  o:     ", dumpTableFew(o,2) )

			local userdata = {
				object     = o,
			}

            if object.properties ~= nil and
               not tableIsEmpty(object.properties)
            then
                userdata.properties = object.properties
            else
                userdata.properties = tile.properties
            end

			o.r = object.rotation or 0
			if o.shape == "rectangle" then
				local cos = math.cos(math.rad(o.r))
				local sin = math.sin(math.rad(o.r))
				local oy  = 0

                -- this is a rectangle
                --
				o.polygon = {
					{ x=o.x+0,   y=o.y+0   },
					{ x=o.x+o.w, y=o.y+0   },
					{ x=o.x+o.w, y=o.y+o.h },
					{ x=o.x+0,   y=o.y+o.h }
				}

                -- apply rotation
                --   change the polygon vertices in-place
                --
				for _, vertex in ipairs(o.polygon) do
					vertex.x, vertex.y = utils.rotate_vertex(map, vertex, o.x, o.y, cos, sin, oy)
				end

				o.vertices = getPolygonVertices(o)
				addObjectToWorld(o, userdata, tile or object)

			elseif o.shape == "ellipse" then
                -- this code is not rotating, polygon will do it
                --
				if not o.polygon then
					o.polygon = utils.convert_ellipse_to_polygon(o.x, o.y, o.w, o.h)
				end
                -- Change it to a polygon and call back in.
                --
                o.shape = "polygon"
                calculateObjectPosition(o, tile)

			elseif o.shape == "polygon" then
				-- Recalculate collision polygons inside tiles
                -- Why?
				if tile then
					local cos = math.cos(math.rad(o.r))
					local sin = math.sin(math.rad(o.r))
					for _, vertex in ipairs(o.polygon) do
						vertex.x = vertex.x + o.x
						vertex.y = vertex.y + o.y
						vertex.x, vertex.y = utils.rotate_vertex(map, vertex, o.x, o.y, cos, sin)
					end
				end

				o.vertices = getPolygonVertices(o)
                addObjectToWorld(o, userdata, tile or object)

			elseif o.shape == "polyline" then
				o.vertices = getPolygonVertices(o)
				addObjectToWorld(o, userdata, tile or object)
			end
		end


        -- finished defining functions
        ------------------------------

        -- process tiles
        --

        -- each entry in the tileInstances table
        -- is a tile gid and a table of places where it appears
        -- (those are the instances)
        --
        -- each instance contains
        --   id      5        ; an instance of a tile
        --   gid     13       ; the tile
        --   layer   table    ; the layer where the tile is used
        --   batch   SpriteBatch
        --   x       128      ; position x
        --   y       0        ; position y
        --   r       0        ; rotation
        --   oy      0        ; not sure what this does
        --
        for gid, instance in pairs(map.tileInstances) do

            local tile = map.tiles[gid]
            print( "\nprocessing this tile" )
            print( dumpTableFew(tile,2) )
            print( "tile properties" )
            print( dumpTableFew(tile.properties,2) )

            -- Every place a tile is used (instanciated)
            --
            for i, place in ipairs(instance) do

                print( "\nprocessing an instance" )
                print( dumpTableFew(place,2) )

                -- there could be more than on object on the tile
                --
                ---- Is the tile's collision polyline stored in
                ---- the object group? Check the map's Lua file.
                --
                if tile.objectGroup then
                    for _, object in ipairs(tile.objectGroup.objects) do
                        print( "\nprocessing a tile object" )
                        print( "gid: ", gid )
                        print( " i : ", i )
                        print( "object:", dumpTableFew(object,2) )
                        print( "object properties:", dumpTableFew(object.properties,2) )
                        if object.properties.collidable == true or
                           tile.properties.collidable == true
                        then
                            object = utils.deepCopy(object)
                            object.dx = place.x + object.x
                            object.dy = place.y + object.y
                            object.width = tile.width
                            object.height = tile.height
                            calculateObjectPosition(object, tile)
                        end
                        print( "" )
                    end
--              else
--                  -- for example, a tile without a collision polyline
--                  print( "tile ", gid, "does not have an objectGroup" )
                end
            end

--          -- Every instance of a tile
--          --
--          if tile.properties.collidable == true then
--              ---- This should use the tile's collision polyline.
--              ---- Right?
--              ----
--              local object = {
--                  shape      = "rectangle",
--                  x          = instance.x,
--                  y          = instance.y,
--                  width      = map.tilewidth,
--                  height     = map.tileheight,
--                  properties = tile.properties
--              }
--              print( "collidable tile object:", dumpTableFew(tile,3) )
--              calculateObjectPosition(object, instance)
--          end
        end

--[[
		for _, layer in ipairs(map.layers) do

			-- Entire layer
            --
			if layer.properties.collidable == true then

				if layer.type == "tilelayer" then
                    print( "tile layer" )
					for gid, tiles in pairs(map.tileInstances) do
						local tile = map.tiles[gid]
						local tileset = map.tilesets[tile.tileset]

						for _, instance in ipairs(tiles) do
							if instance.layer == layer then
								local object = {
									shape      = "rectangle",
									x          = instance.x,
									y          = instance.y,
									width      = tileset.tilewidth,
									height     = tileset.tileheight,
									properties = tile.properties
								}
                                print( "layer object:", dumpTableFew(object,3) )

								calculateObjectPosition(object, instance)
							end
						end
					end

				elseif layer.type == "objectgroup" then
                    print( "objectgroup layer" )
					for _, object in ipairs(layer.objects) do
						calculateObjectPosition(object)
					end

				elseif layer.type == "imagelayer" then
					local object = {
						shape      = "rectangle",
						x          = layer.x or 0,
						y          = layer.y or 0,
						width      = layer.width,
						height     = layer.height,
						properties = layer.properties
					}
                    print( "image layer:", dumpTableFew(object,3) )

					calculateObjectPosition(object)
				end
			end

			-- Individual objects
			if layer.type == "objectgroup" then
				for _, object in ipairs(layer.objects) do
					if object.properties.collidable == true then
						calculateObjectPosition(object)
					end
				end
			end
		end
--]]

		map.box2d_collision = collision
	end,

	--- Remove Box2D fixtures and shapes from world.
	-- @param index The index or name of the layer being removed
	box2d_removeLayer = function(map, index)
		local layer = assert(map.layers[index], "Layer not found: " .. index)
		local collision = map.box2d_collision

		-- Remove collision objects
		for i = #collision, 1, -1 do
			local obj = collision[i]

			if obj.object.layer == layer then
				obj.fixture:destroy()
				table.remove(collision, i)
			end
		end
	end,

	--- Draw Box2D physics world.
	-- @param tx Translate on X
	-- @param ty Translate on Y
	-- @param sx Scale on X
	-- @param sy Scale on Y
	box2d_draw = function(map, tx, ty, sx, sy)
		local collision = map.box2d_collision

		lg.push()
		lg.scale(sx or 1, sy or sx or 1)
		lg.translate(math.floor(tx or 0), math.floor(ty or 0))

		for _, obj in ipairs(collision) do
			local points = {obj.body:getWorldPoints(obj.shape:getPoints())}
			local shape_type = obj.shape:getType()

			if shape_type == "edge" or shape_type == "chain" then
				love.graphics.line(points)
			elseif shape_type == "polygon" then
				love.graphics.polygon("line", points)
			else
				error("sti box2d plugin does not support "..shape_type.." shapes")
			end
		end

		lg.pop()
	end
}

--- Custom Properties in Tiled are used to tell this plugin what to do.
-- @table Properties
-- @field collidable set to true, can be used on any Layer, Tile, or Object
-- @field sensor set to true, can be used on any Tile or Object that is also collidable
-- @field dynamic set to true, can be used on any Tile or Object
-- @field friction can be used to define the friction of any Object
-- @field restitution can be used to define the restitution of any Object
-- @field categories can be used to set the filter Category of any Object
-- @field mask can be used to set the filter Mask of any Object
-- @field group can be used to set the filter Group of any Object
